"""
Methods for using the 'tag' command
"""
import re
import subprocess
from pathlib import Path
from typing import Optional, Union

from .constants import TAG_ALREADY_EXISTS_RE, TAG_NOT_FOUND_RE
from .exceptions import (AlreadyExistsException, DoesNotExistException,
                         GitException)

__all__ = [
    "list_tags", "create_tag",
    "delete_tag",
]


def list_tags(git_repo: Union[Path, str], tag_pattern: Optional[str] = None) -> list[str]:
    """
    List all git tags or filter with a wildcard pattern

        :param git_repo: Path to the repo
        :param tag_pattern: Filter the tag list with a wildcard pattern, defaults to None
        :raises GitException: Error to do with git
        :return: List of found git tags
    """
    args = ["git", "-C", str(git_repo), "tag", "-l"]
    if tag_pattern:
        args.append(tag_pattern)

    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        raise GitException(stderr)

    return process_status.stdout.decode().strip().split("\n")


def create_tag(git_repo: Union[Path, str], tag_name: str, commit_hash: Optional[str] = None):
    """
    Create a new lightweight tag

        :param git_repo: Path to the repo
        :param tag_name: The tag name to use
        :param commit_hash: Create tag on a different
                            commit other than HEAD, defaults to None
        :raises AlreadyExistsException: When the tag name already exists
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "tag", tag_name]
    if commit_hash:
        args.append(commit_hash)

    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(TAG_ALREADY_EXISTS_RE, stderr):
            raise AlreadyExistsException(f"tag '{tag_name}' already exists")
        raise GitException(stderr)


def delete_tag(git_repo: Union[Path, str], tag_name: str) -> str:
    """
    Delete a tag

        :param git_repo: Path to the repo
        :param tag_name: The tag name to use
        :raises DoesNotExistException: The tag was not found
        :raises GitException: Error to do with git
        :return: Output provided by the git when a tag is removed
    """
    args = ["git", "-C", str(git_repo), "tag", "-d", tag_name]

    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(TAG_NOT_FOUND_RE, stderr):
            raise DoesNotExistException(f"tag '{tag_name}' not found")
        raise GitException(stderr)

    return process_status.stdout.decode().strip().removesuffix("\n")



"""
Methods for using the 'log' command
"""
import re
import subprocess
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Optional

from .constants import EMPTY_REPO_RE, UNKNOWN_REV_RE
from .datatypes import Log
from .exceptions import (GitException, NoCommitsException, NoLogsException,
                         UnknownRevisionException)

__all__ = ["get_logs"]


def __process_log(stdout_line: str) -> Log:
    parts = stdout_line.split(";;")
    if len(parts) != 6:
        raise ValueError(f"invalid log line: {stdout_line}")
    return Log(
        parts[0],
        parts[1],
        parts[2],
        parts[3],
        datetime.fromisoformat(parts[4]),
        parts[5],
    )


def __process_logs(stdout: str) -> Iterator[Log]:
    log_lines = stdout.strip().split("\n")
    return map(__process_log, log_lines)


def get_logs(
        git_repo: Path,
        branch: Optional[str] = None,
        max_number: Optional[int] = None,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None) -> Iterator[Log]:
    """
    Generate git logs from a repo

        :param git_repo: Path to the repo
        :param branch: The branch name, defaults to None
        :param max_number: max number of logs to get, defaults to None
        :param since: Filter logs after given date, defaults to None
        :param until: Filter logs before given date defaults to None
        :raises NoCommitsException: Repo has no commits
        :raises UnknownRevisionException: Unknown revision/branch name
        :raises GitException: Error to do with git
        :raises NoLogsException: No logs have been generated
        :return: The generated logs
    """
    args = ["git", "-C", str(git_repo), "log"]

    if branch is not None:
        args.append(str(branch))
    if max_number is not None:
        args.append(f"--max-count={max_number}")
    if since is not None:
        args.append(f"--since={since.isoformat()}")
    if until is not None:
        args.append(f"--until={until.isoformat()}")
    # formats: https://git-scm.com/docs/pretty-formats
    args.append("--pretty=%H;;%P;;%ae;;%an;;%cI;;%s")

    process_status = subprocess.run(
        args,
        capture_output=True)
    if not process_status.stdout:
        stderr = process_status.stderr.decode()
        if re.match(EMPTY_REPO_RE, stderr):
            raise NoCommitsException()
        if re.match(UNKNOWN_REV_RE, stderr):
            raise UnknownRevisionException(f"unknown revision/branch {branch}")
        if process_status.returncode != 0:
            raise GitException(stderr)
        raise NoLogsException(f"no logs found (using given filters) for '{git_repo.name}'")
    stdout = process_status.stdout.decode()
    return __process_logs(stdout)



"""
Methods for using the 'cat-file' command
"""
import re
import subprocess
from pathlib import Path
from typing import Union

from .constants import NOT_VALID_OBJECT_NAME_RE
from .datatypes import TreeContentTypes
from .exceptions import GitException, UnknownRevisionException

__all__ = ["get_object_size", "get_object_type", "get_pretty_print"]


def __cat_file_command(git_repo: Union[Path, str], tree_ish: str, file_path: str, *flags) -> bytes:
    args = ["git", "-C", str(git_repo), "cat-file", f"{tree_ish}:{file_path}"]
    args.extend(flags)

    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(NOT_VALID_OBJECT_NAME_RE, stderr):
            raise UnknownRevisionException(f"Invalid object name '{tree_ish}:{file_path}'")
        raise GitException(stderr)

    return process_status.stdout


def get_object_size(git_repo: Union[Path, str], tree_ish: str, file_path: str) -> int:
    """
    Gets the objects size from repo

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param file_path: The file in the repo to read
        :raises UnknownRevisionException: Invalid tree_ish or file_path
        :raises GitException: Error to do with git
        :return: The object size
    """
    return int(__cat_file_command(git_repo, tree_ish, file_path, "-s"))


def get_object_type(git_repo: Union[Path, str], tree_ish: str, file_path: str) -> TreeContentTypes:
    """
    Gets the object type from repo

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param file_path: The file in the repo to read
        :raises UnknownRevisionException: Invalid tree_ish or file_path
        :raises GitException: Error to do with git
        :return: The object type
    """
    output = __cat_file_command(git_repo, tree_ish, file_path, "-t").decode()
    return TreeContentTypes(output)


def get_pretty_print(git_repo: Union[Path, str], tree_ish: str, file_path: str) -> bytes:
    """
    Gets a object from repo

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param file_path: The file in the repo to read
        :raises UnknownRevisionException: Invalid tree_ish or file_path
        :raises GitException: Error to do with git
        :return: The object type
    """
    return __cat_file_command(git_repo, tree_ish, file_path, "-p")



"""
Methods not related to git commands,
but to help the program function
"""
from pathlib import Path
from typing import Union


def ensure_path(path_or_str: Union[Path, str]) -> Path:
    """
    Ensures that given value is a pathlib.Path object.

        :param path_or_str: Either a path string or pathlib.Path obj
        :return: The value as a pathlib.Path obj
    """
    return path_or_str if isinstance(path_or_str, Path) else Path(path_or_str)



"""
Methods for using the 'show' command
"""
import re
import subprocess
from pathlib import Path
from typing import Generator, Union

from .constants import INVALID_OBJECT_NAME, PATH_DOES_NOT_EXIST
from .exceptions import (GitException, PathDoesNotExistInRevException,
                         UnknownRevisionException)

__all__ = [
    "show_file", "show_file_buffered",
]


def show_file(git_repo: Union[Path, str], tree_ish: str, file_path: str) -> bytes:
    """
    Read a file from a repository

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param file_path: The file in the repo to read
        :raises UnknownRevisionException: Unknown tree_ish
        :raises PathDoesNotExistInRevException: File not found in repo
        :raises GitException: Error to do with git
        :return: The read file
    """
    args = ["git", "-C", str(git_repo), "show", f"{tree_ish}:{file_path}"]

    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(INVALID_OBJECT_NAME, stderr):
            raise UnknownRevisionException(f"Unknown tree-ish '{tree_ish}'")
        elif re.match(PATH_DOES_NOT_EXIST, stderr):
            raise PathDoesNotExistInRevException(f"'{file_path}' not found in repo")
        raise GitException(stderr)

    return process_status.stdout


def show_file_buffered(
        git_repo: Union[Path, str],
        tree_ish: str,
        file_path: str,
        bufsize: int = -1) -> Generator[bytes, None, None]:
    """
    Read a file from a repository, but using a buffered read

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param file_path: The file in the repo to read
        :param bufsize: The buffer size to pass into subprocess.Popen, defaults to -1
        :raises UnknownRevisionException: Unknown tree_ish
        :raises PathDoesNotExistInRevException: File not found in repo
        :raises GitException: Error to do with git
        :yield: Each read file section
    """
    args = ["git", "-C", str(git_repo), "show", f"{tree_ish}:{file_path}"]

    with subprocess.Popen(
        args,
        bufsize=bufsize,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    ) as process:
        for line in process.stdout:
            yield line

        return_code = process.wait()
        if return_code != 0:
            stderr = process.stderr.read().decode()
            if re.match(INVALID_OBJECT_NAME, stderr):
                raise UnknownRevisionException(f"Unknown tree-ish '{tree_ish}'")
            elif re.match(PATH_DOES_NOT_EXIST, stderr):
                raise PathDoesNotExistInRevException(f"'{file_path}' not found in repo")
            raise GitException(stderr)



"""
Methods for using the 'rev-list' command
"""
import re
import subprocess
from pathlib import Path
from typing import Optional, Union

from .constants import UNKNOWN_REV_RE
from .exceptions import GitException, UnknownRevisionException

__all__ = [
    "get_commit_count", "get_disk_usage",
    "get_rev_list",
]


def _rev_list(
        git_repo: Union[Path, str],
        branch: Optional[str] = None,
        operator: Optional[str] = None) -> str:
    if branch is None:
        branch = "--all"
    args = ["git", "-C", str(git_repo), "rev-list", branch]
    if operator:
        args.append(operator)
    process_status = subprocess.run(args, capture_output=True)

    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(UNKNOWN_REV_RE, stderr):
            raise UnknownRevisionException(stderr)
        raise GitException(stderr)
    return process_status.stdout.decode()


def get_commit_count(git_repo: Union[Path, str], branch: Optional[str] = None) -> int:
    """
    Get a repos commit count

        :param git_repo: Path to the repo
        :param branch: Branch to filter, defaults to None
        :raises UnknownRevisionException: Unknown tree_ish
        :raises GitException: Error to do with git
        :return: The commit count
    """
    return int(_rev_list(git_repo, branch, "--count"))


def get_disk_usage(git_repo: Union[Path, str], branch: Optional[str] = None) -> int:
    """
    Get a size of the repo

        :param git_repo: Path to the repo
        :param branch: Branch to filter, defaults to None
        :raises UnknownRevisionException: Unknown tree_ish
        :raises GitException: Error to do with git
        :return: The size of the repo
    """
    return int(_rev_list(git_repo, branch, "--disk-usage"))


def get_rev_list(git_repo: Union[Path, str], branch: Optional[str] = None) -> list[str]:
    """
    Get a repos revisions

        :param git_repo: Path to the repo
        :param branch: Branch to filter, defaults to None
        :raises UnknownRevisionException: Unknown tree_ish
        :raises GitException: Error to do with git
        :return: The repos revisions
    """
    return _rev_list(git_repo, branch).strip().split("\n")



"""
Methods that don't fit in their own file
"""
import os
import subprocess
from pathlib import Path
from typing import Generator, Optional, Union

from .datatypes import ArchiveTypes
from .exceptions import AlreadyExistsException, GitException
from .helpers import ensure_path

__all__ = [
    "get_version", "init_repo",
    "clone_repo", "get_description",
    "set_description", "run_maintenance",
    "get_archive", "get_archive_buffered",
]


def get_version() -> str:
    """
    Gets the git version

        :raises GitException: Error to do with git
        :return: The version
    """
    args = ("git", "version")
    process_status = subprocess.run(args, capture_output=True)
    if process_status.returncode != 0:
        raise GitException(process_status.stderr.decode())
    return process_status.stdout.decode().strip()


def init_repo(
        repo_dir: Path,
        repo_name: str,
        bare: bool = True,
        default_branch: Optional[str] = None):
    """
    Creates a new git repo in the directory with the given name,
    if bare the repo name will have .git added at the end.

        :param repo_dir: Where the repo will be
        :param repo_name: The name of the repo
        :param bare: Whether the repo is bare, defaults to True
        :param default_branch: The branch name to use, defaults to None
        :raises AlreadyExistsException: A repo already exists
        :raises GitException: Error to do with git
    """
    if bare:
        repo_name = repo_name + ".git"
    repo_path = repo_dir / repo_name

    if (repo_path.exists()):
        raise AlreadyExistsException(f"path already exists for '{repo_name}'")

    args = ["git", "init", str(repo_path), "--quiet"]
    if bare:
        args.append("--bare")
    if default_branch:
        args.append(f"--initial-branch={default_branch}")
    process = subprocess.run(args)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())


def clone_repo(git_repo: Union[Path, str], src: str, bare=False, mirror=False):
    """
    Clone an exiting repo, please note this
    method has no way of passing passwords+usernames

        :param git_repo: Repo path to clone into
        :param src: Where to clone from
        :param bare: Use --bare git argument, defaults to False
        :param mirror: Use --mirror git argument, defaults to False
        :raises ValueError: Both bare and mirror are True
        :raises GitException: Error to do with git
    """
    args = ["git", "clone", src, str(git_repo)]

    # disables interactive password prompt
    env = {
        **os.environ,
        "GCM_INTERACTIVE": "never",
        "GIT_TERMINAL_PROMPT": "0"
    }

    if mirror and bare:
        raise ValueError("both bare and mirror cannot be used at same time")
    elif bare:
        args.append("--bare")
    elif mirror:
        args.append("--mirror")

    process = subprocess.run(args, capture_output=True, env=env)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())


def get_description(git_repo: Union[Path, str]) -> str:
    """
    Gets the set description for a repo

        :param git_repo: Path to the repo
        :return: The description
    """
    git_repo = ensure_path(git_repo)
    with open(git_repo / "description", "r") as fo:
        return fo.read()


def set_description(git_repo: Union[Path, str], description: str):
    """
    Sets the set description for a repo

        :param git_repo: Path to the repo
    """
    git_repo = ensure_path(git_repo)
    with open(git_repo / "description", "w") as fo:
        fo.write(description)


def run_maintenance(git_repo: Union[Path, str]):
    """
    Run a maintenance git command to specified repo

        :param git_repo: Where the repo is
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "maintenance", "run"]
    process = subprocess.run(args, capture_output=True)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())


def get_archive(
        git_repo: Union[Path, str],
        archive_type: ArchiveTypes,
        tree_ish: str = "HEAD") -> bytes:
    """
    get a archive of a git repo

        :param git_repo: Where the repo is
        :param archive_type: What archive type will be created
        :param tree_ish: What commit/branch to save, defaults to "HEAD"
        :raises GitException: Error to do with git
        :return: The content of the archive ready to write to a file
    """
    # this allows for strings to be passed
    if isinstance(archive_type, ArchiveTypes):
        archive_type = archive_type.value
    process = subprocess.run(
        ["git", "-C", str(git_repo), "archive", f"--format={archive_type}", tree_ish],
        capture_output=True)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())
    return process.stdout


def get_archive_buffered(
        git_repo: Union[Path, str],
        archive_type: ArchiveTypes,
        tree_ish: str = "HEAD",
        bufsize: int = -1) -> Generator[bytes, None, None]:
    """
    get a archive of a git repo, but using a buffered read

        :param git_repo: Where the repo is
        :param archive_type: What archive type will be created
        :param tree_ish: What commit/branch to save, defaults to "HEAD"
        :param bufsize: The buffer size to pass into subprocess.Popen, defaults to -1
        :raises GitException: Error to do with git
        :yield: Each read content section
    """
    # this allows for strings to be passed
    if isinstance(archive_type, ArchiveTypes):
        archive_type = archive_type.value

    with subprocess.Popen(
        ["git", "-C", str(git_repo), "archive", f"--format={archive_type}", tree_ish],
        bufsize=bufsize,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    ) as process:
        for line in process.stdout:
            yield line

        return_code = process.wait()
        if return_code != 0:
            stderr = process.stderr.read().decode()
            raise GitException(stderr)



"""
Exceptions that could be raised during one of the git commands
"""


class GitException(Exception):
    """
    Parent exception for all git exceptions,
    used when there is no other exception that fits error
    """
    pass


class NoCommitsException(GitException):
    """
    Raised when a repository has commits
    """
    pass


class NoBranchesException(GitException):
    """
    Raised when a repository has no branches
    or none that match a filter
    """
    pass


class AlreadyExistsException(GitException):
    """
    Raised when something already exists
    could be repository, branch, etc
    """
    pass


class DoesNotExistException(GitException):
    """
    Raised when something does not exist e.g. tag
    """
    pass


class NoLogsException(GitException):
    """
    Raised when a repository has no logs available
    """
    pass


class UnknownRevisionException(GitException):
    """
    Raised when a revision is not found
    """
    pass


class UnknownRefException(GitException):
    """
    Raised when a a reference is not found
    """
    pass


class PathDoesNotExistInRevException(GitException):
    """
    Raised when a path does not exist in a repository
    """
    pass
